module Labyrinth where

-- Modelling Labyrinths: a maze of twisty little passages, all alike

-- An identifier for the nodes:
type Id  = Integer

-- The passageways in the labyrinth (nodes in the graph):
data Lab = Dead Id 
         | Pass Id Lab 
         | TJnc Id Lab Lab

-- Get identity of a node
nid :: Lab-> Id
nid (Dead i) = i
nid (Pass i _) = i
nid (TJnc i _ _)= i

-- A path in the labyrinth (essentially, a list of node identifiers):
data Path = Cons Id Path
          | Mt
          deriving (Eq, Show)

-- Checks wether a node is contained in a path
contains :: Id-> Path-> Bool
contains _ Mt = False
contains i (Cons j js) = i == j || contains i js

-- Add at the back
snoc :: Path-> Id-> Path
snoc (Cons i p) j = Cons i (snoc p j)
snoc Mt         j = Cons j Mt

-- A traversal: a successful path, or a failure:
data Trav = Succ Path
          | Fail
          deriving (Eq, Show)

-- Add a node to a traversal, if successful:
cons :: Id-> Trav-> Trav
cons _ Fail      = Fail
cons i (Succ is) = Succ (Cons i is)

-- Add two traversals, disregarding first if it fails:
select :: Trav-> Trav-> Trav
select Fail t = t
select t    _ = t

-- Traversal of a cycle-free labyrinth to a given node, simplest version.
traverse1 :: Id-> Lab-> Trav
traverse1 t l
  | nid l == t = Succ (Cons (nid l) Mt)
  | otherwise = case l of
    Dead _ -> Fail
    Pass i n -> cons i (traverse1 t n)
    TJnc i n m -> select (cons i (traverse1 t n)) 
                         (cons i (traverse1 t m))

-- Traversal of a labyrinth with cycles to a given node.
traverse2 :: Id-> Lab-> Path-> Trav
traverse2 t l p
  | nid l == t = Succ (snoc p (nid l))
  | contains (nid l) p = Fail
  | otherwise = case l of
    Dead _ -> Fail
    Pass i n -> traverse2 t n (snoc p i)
    TJnc i n m -> select (traverse2 t n (snoc p i))
                         (traverse2 t m (snoc p i))


-- An example labyrinth without cycles:
l00 = Pass 0 l01
l10 = Pass 1 l00   -- "Entrance"
l20 = Pass 2 l21
l30 = Pass 3 l20
l40 = Dead 4
l01 = Pass 5 l02 
l11 = Dead 6
l21 = Pass 7 l11
l31 = TJnc 8 l30 l41
l41 = TJnc 9 l40 l42
l02 = Pass 10 l12
l12 = Pass 11 l13
l22 = Pass 12 l32
l32 = Pass 13 l31
l42 = Dead 14
l03 = Dead 15
l13 = Pass 16 l23 -- Dead 16 -- This change bisects the labyrinth
l23 = TJnc 17 l22 l24 
l33 = Pass 18 l43
l43 = Pass 19 l44
l04 = Pass 20 l03
l14 = Pass 21 l04
l24 = TJnc 22 l14 l34
l34 = Pass 23 l33
l44 = Dead 24     -- "Exit"

-- An example labyrinth with cycles:
m00 = Pass 0 m01
m10 = Pass 1 m00  -- "Entrance"
m20 = Pass 2 m21
m30 = Pass 3 m20
m40 = Dead 4
m01 = Pass 5 m02 
m11 = Pass 6 m01 -- This makes one cycle
m21 = Pass 7 m11
m31 = TJnc 8 m30 m41
m41 = TJnc 9 m40 m42
m02 = Pass 10 m12
m12 = Pass 11 m13
m22 = Pass 12 m32
m32 = Pass 13 m31
m42 = Dead 14
m03 = Pass 15 m02 -- This makes another cycle
m13 = Pass 16 m23
m23 = TJnc 17 m22 m24 
m33 = Pass 18 m43
m43 = Pass 19 m44
m04 = Pass 20 m03
m14 = Pass 21 m04 
m24 = TJnc 22 m14 m34
m34 = Pass 23 m33
m44 = Dead 24

